@@ -1,9 +1,16 @@
 #' Capture a backtrace
 #'
+#' @description
 #' A backtrace captures the sequence of calls that lead to the current
-#' function, sometimes called the call stack. Because of lazy
+#' function (sometimes called the call stack). Because of lazy
 #' evaluation, the call stack in R is actually a tree, which the
-#' `summary()` method of this object will reveal.
+#' `print()` method for this object will reveal.
+#'
+#' Users rarely need to call `trace_back()` manually. Instead,
+#' signalling an error with [abort()] or setting up [global_entrace()]
+#' is the most common way to create backtraces when an error is
+#' thrown. Inspect the backtrace created for the most recent error
+#' with [last_error()].
 #'
 #' `trace_length()` returns the number of frames in a backtrace.
 #'
@@ -15,6 +22,13 @@
 #'   indirectly or from a larger context, for example in tests or
 #'   inside an RMarkdown document where you don't want all of the
 #'   knitr evaluation mechanisms to appear in the backtrace.
+#'
+#'   If not supplied, the `rlang_trace_top_env` global option is
+#'   consulted. This makes it possible to trim the embedding context
+#'   for all backtraces created while the option is set. If knitr is
+#'   in progress, the default value for this option is
+#'   `knitr::knit_global()` so that the knitr context is trimmed out
+#'   of backtraces.
 #' @param bottom The last frame environment to be included in the
 #'   backtrace. This becomes the rightmost leaf of the backtrace tree
 #'   and represents the youngest call in the backtrace.
@@ -76,27 +90,46 @@
 #' close(conn)
 #' @export
 trace_back <- function(top = NULL, bottom = NULL) {
+  # FIXME: Include this in the `trace_back()` UI?
+  visible_bottom <- peek_option("rlang:::visible_bottom")
+
   frames <- sys.frames()
+
   idx <- trace_find_bottom(bottom, frames)
+  visible_idx <- trace_find_bottom(visible_bottom, frames)
+
+  visible_idx <- intersect(visible_idx, idx)
+  is_visible <- seq_along(idx) %in% visible_idx
 
   frames <- frames[idx]
   parents <- sys.parents()[idx]
   calls <- as.list(sys.calls()[idx])
 
   calls <- map(calls, call_fix_car)
-  calls <- add_pipe_pointer(calls, frames)
-  calls <- map2(calls, seq_along(calls), maybe_add_namespace)
+  calls <- map(calls, call_zap_inline)
+
+  context <- map2(calls, seq_along(calls), call_trace_context)
+  context <- inject(vec_rbind(empty_trace_context(), !!!context))
 
   parents <- normalise_parents(parents)
 
-  trace <- new_trace(calls, parents)
+  trace <- new_trace(
+    calls,
+    parents,
+    namespace = context$namespace,
+    scope = context$scope,
+    visible = is_visible
+  )
   trace <- add_winch_trace(trace)
   trace <- trace_trim_env(trace, frames, top)
 
   trace
 }
 
-trace_find_bottom <- function(bottom, frames) {
+trace_find_bottom <- function(bottom,
+                              frames,
+                              arg = caller_arg(bottom),
+                              call = caller_env()) {
   if (is_null(bottom)) {
     return(seq_len(sys.parent(2L)))
   }
@@ -107,17 +140,36 @@ trace_find_bottom <- function(bottom, frames) {
       if (is_reference(bottom, global_env())) {
         return(int())
       }
-      abort("Can't find `bottom` on the call tree")
+      msg <- sprintf(
+        "Can't find %s on the call tree.",
+        format_arg(arg)
+      )
+      abort(msg, call = call)
     }
 
     return(seq_len(top))
   }
 
   if (is_integerish(bottom, n = 1)) {
-    return(seq_len(sys.parent(bottom + 1L)))
+    if (bottom < 0) {
+      msg <- sprintf(
+        "%s must be a positive integer.",
+        format_arg(arg)
+      )
+      abort(msg, call = call)
+    }
+    if (inherits(bottom, "AsIs")) {
+      return(seq_len(bottom))
+    } else {
+      return(seq_len(sys.parent(bottom + 1L)))
+    }
   }
 
-  abort("`bottom` must be `NULL`, a frame environment, or an integer")
+  msg <- sprintf(
+    "%s must be `NULL`, a frame environment, or an integer.",
+    format_arg(arg)
+  )
+  abort(msg, call = call)
 }
 
 # Work around R bug causing promises to leak in frame calls
@@ -128,81 +180,34 @@ call_fix_car <- function(call) {
   call
 }
 
-# Assumes magrittr 1.5
-add_pipe_pointer <- function(calls, frames) {
-  pipe_begs <- which(map_lgl(calls, is_call2, "%>%"))
-  pipe_kinds <- map_int(pipe_begs, pipe_call_kind, calls)
-
-  pipe_calls <- map2(pipe_begs, pipe_kinds, function(beg, kind) {
-    call <- calls[[beg]]
-
-    if (kind == 0L) {
-      return(call)
-    }
-
-    if (kind == 1L) {
-      v <- "i"
-    } else if (kind == 2L) {
-      v <- "k"
-    }
-
-    frame <- frames[[beg + 5L]]
-    pointer <- frame[[v]]
-
-    fn <- frame[["function_list"]][[1]]
-    info <- pipe_collect_calls(call, fn_env(fn))
-
-    structure(call, pipe_pointer = pointer, pipe_info = info)
-  })
-
-  calls[pipe_begs] <- pipe_calls
-  calls
-}
-
-pipe_call_kind <- function(beg, calls) {
-  end1 <- beg + 6L
-  end2 <- beg + 7L
-
-  if (end2 > length(calls)) {
-    return(0L)
-  }
-
-  # Uncomplete pipe call
-  magrittr_call1 <- quote(function_list[[i]](value))
-
-  # Last call of the pipe
-  magrittr_call2 <- quote(function_list[[k]](value))
-
-  if (identical(calls[[end1]], magrittr_call1)) {
-    return(1L)
-  }
-  if (identical(calls[[end2]], magrittr_call2)) {
-    return(2L)
-  }
-  0L
-}
-
-maybe_add_namespace <- function(call, fn) {
+call_trace_context <- function(call, fn) {
   if (is_quosure(call)) {
     call <- quo_get_expr(call)
     if (!is_call(call)) {
-      return(call)
+      return(trace_context())
     }
   }
 
   if (call_print_fine_type(call) != "call") {
-    return(call)
+    return(trace_context())
   }
 
-  # Checking for bare symbols covers the `::` and `:::` cases
-  sym <- node_car(call)
-  if (!is_symbol(sym)) {
-    return(call)
+  namespace <- call_ns(call)
+  name <- call_name(call)
+
+  if (is_null(name)) {
+    return(trace_context())
+  }
+
+  if (!is_null(namespace)) {
+    return(trace_context(
+      namespace = namespace,
+      scope = as_string(call[[1]][[1]])
+    ))
   }
 
-  nm <- as_string(sym)
   if (is_environment(fn)) {
-    fn <- get(nm, envir = fn, mode = "function")
+    fn <- get(name, envir = fn, mode = "function")
   } else if (is_function(fn)) {
     fn <- fn
   } else {
@@ -212,22 +217,36 @@ maybe_add_namespace <- function(call, fn) {
   env <- fn_env(fn)
   top <- topenv(env)
   if (is_reference(env, global_env())) {
-    prefix <- "global"
-    op <- "::"
+    namespace <- NA
+    scope <- "global"
   } else if (is_namespace(top)) {
-    prefix <- ns_env_name(top)
-    if (ns_exports_has(top, nm)) {
-      op <- "::"
+    namespace <- ns_env_name(top)
+    if (ns_exports_has(top, name)) {
+      scope <- "::"
+    } else if (env_has(top, name)) {
+      scope <- ":::"
     } else {
-      op <- ":::"
+      scope <- "local"
     }
   } else {
-    return(call)
+    namespace <- NA
+    scope <- NA
   }
 
-  namespaced_sym <- call(op, sym(prefix), sym)
-  call[[1]] <- namespaced_sym
-  call
+  trace_context(
+    namespace = namespace,
+    scope = scope
+  )
+}
+
+trace_context <- function(namespace = NA, scope = NA) {
+  data_frame(
+    namespace = namespace,
+    scope = scope
+  )
+}
+empty_trace_context <- function() {
+  trace_context(chr(), chr())
 }
 
 # Remove recursive frames which occur with quosures
@@ -237,33 +256,6 @@ normalise_parents <- function(parents) {
   parents
 }
 
-new_trace <- function(calls, parents, indices = NULL) {
-  indices <- indices %||% seq_along(calls)
-
-  n <- length(calls)
-  stopifnot(
-    is_list(calls),
-    is_integer(parents, n),
-    is_integer(indices, n)
-  )
-
-  structure(
-    list(
-      calls = calls,
-      parents = parents,
-      indices = indices
-    ),
-    class = "rlang_trace",
-    # Increment this number when the internal format for the class changes
-    version = 1L
-  )
-}
-
-trace_reset_indices <- function(trace) {
-  trace$indices <- seq_len(trace_length(trace))
-  trace
-}
-
 # Can't use new_environment() here
 winch_available_env <- new.env(parent = emptyenv())
 
@@ -286,17 +278,114 @@ add_winch_trace <- function(trace) {
   winch::winch_add_trace_back(trace)
 }
 
+
+# Construction ------------------------------------------------------------
+
+new_trace <- function(call,
+                      parent,
+                      ...,
+                      visible = TRUE,
+                      namespace = na_chr,
+                      scope = na_chr,
+                      class = NULL) {
+  new_trace0(
+    call,
+    parent,
+    ...,
+    visible = visible,
+    namespace = namespace,
+    scope = scope,
+    class = c(class, "rlang_trace", "rlib_trace")
+  )
+}
+new_trace0 <- function(call,
+                       parent,
+                       ...,
+                       visible = TRUE,
+                       namespace = NA,
+                       scope = NA,
+                       class = NULL) {
+  if (is_pairlist(call)) {
+    call <- as.list(call)
+  }
+  stopifnot(
+    is_bare_list(call),
+    is_bare_integer(parent)
+  )
+
+  df <- df_list(
+    call = call,
+    parent = parent,
+    visible = visible,
+    namespace = namespace,
+    scope = scope,
+    ...
+  )
+  new_data_frame(
+    df,
+    .class = c(class, "tbl"),
+    version = 2L
+  )
+}
+
+
+# Operations --------------------------------------------------------------
+
+#' @rdname trace_back
+#' @param trace A backtrace created by `trace_back()`.
+#' @export
+trace_length <- function(trace) {
+  nrow(trace)
+}
+
+trace_slice <- function(trace, i) {
+  i <- vec_as_location(i, trace_length(trace))
+
+  parent <- match(trace$parent, i, nomatch = 0)
+
+  out <- vec_slice(trace, i)
+  out$parent <- parent[i]
+
+  out
+}
+
+trace_bind <- function(...) {
+  traces <- compact(list2(...))
+  n <- length(traces)
+
+  if (!every(traces, inherits, "rlang_trace")) {
+    abort("`...` must contain backtraces.")
+  }
+
+  if (n == 0L) {
+    return(new_trace(list(), int()))
+  }
+  if (n == 1L) {
+    return(traces[[1]])
+  }
+
+  out <- reduce(traces, function(x, y) {
+    if (identical(x$call, y$call)) {
+      return(x)
+    }
+    y$parent <- y$parent + nrow(x)
+    vec_rbind(as.data.frame(x), as.data.frame(y))
+  })
+
+  new_data_frame(out, .class = c("rlang_trace", "rlib_trace", "tbl"))
+}
+
+
 # Methods -----------------------------------------------------------------
 
 # For internal use only
 c.rlang_trace <- function(...) {
   traces <- list(...)
 
-  calls <- flatten(map(traces, `[[`, "calls"))
-  parents <- flatten_int(map(traces, `[[`, "parents"))
-  indices <- flatten_int(map(traces, `[[`, "indices"))
+  calls <- flatten(map(traces, `[[`, "call"))
+  parents <- flatten_int(map(traces, `[[`, "parent"))
 
-  new_trace(calls, parents, indices)
+  new_trace(calls, parents)
 }
 
 #' @export
@@ -306,10 +395,9 @@ format.rlang_trace <- function(x,
                                max_frames = NULL,
                                dir = getwd(),
                                srcrefs = NULL) {
-  x <- trace_reset_indices(x)
-
-  switch(arg_match(simplify),
-    none = trace_format(x, max_frames, dir, srcrefs),
+  switch(
+    arg_match(simplify),
+    none = trace_format_full(x, max_frames, dir, srcrefs),
     collapse = trace_format_collapse(x, max_frames, dir, srcrefs),
     branch = trace_format_branch(x, max_frames, dir, srcrefs)
   )
@@ -324,8 +412,14 @@ trace_format <- function(trace, max_frames, dir, srcrefs) {
     return(trace_root())
   }
 
-  tree <- trace_as_tree(trace, dir = dir, srcrefs = srcrefs)
-  cli_tree(tree, indices = trace$indices)
+  cli_tree(trace_as_tree(trace, dir = dir, srcrefs = srcrefs))
+}
+
+trace_format_full <- function(trace, max_frames, dir, srcrefs) {
+  if (length(trace$visible)) {
+    trace$visible <- TRUE
+  }
+  trace_format(trace, max_frames, dir, srcrefs)
 }
 trace_format_collapse <- function(trace, max_frames, dir, srcrefs) {
   trace <- trace_simplify_collapse(trace)
@@ -333,11 +427,12 @@ trace_format_collapse <- function(trace, max_frames, dir, srcrefs) {
 }
 trace_format_branch <- function(trace, max_frames, dir, srcrefs) {
   trace <- trace_simplify_branch(trace)
-  trace <- branch_uncollapse_pipe(trace)
   tree <- trace_as_tree(trace, dir = dir, srcrefs = srcrefs)
 
-  branch <- tree[-1, ][["call"]]
-  cli_branch(branch, max = max_frames, indices = trace$indices)
+  # Remove root in the branch view
+  tree <- vec_slice(tree, -1)
+
+  cli_branch(tree, max = max_frames)
 }
 
 format_collapsed <- function(what, n) {
@@ -351,27 +446,28 @@ format_collapsed <- function(what, n) {
 
   paste0(what, n_text)
 }
-format_collapsed_branch <- function(what, n, style = NULL) {
-  style <- style %||% cli_box_chars()
-  what <- sprintf(" %s %s", style$h, what)
-  format_collapsed(what, n)
-}
 
-cli_branch <- function(lines, max = NULL, style = NULL, indices = NULL) {
+cli_branch <- function(tree,
+                       max = NULL,
+                       style = NULL) {
+  lines <- tree$call_text
+
   if (!length(lines)) {
     return(chr())
   }
 
-  numbered <- length(indices)
-  if (numbered) {
-    indices <- pad_spaces(as.character(indices))
-    indices <- paste0(" ", indices, ". ")
-    padding <- nchar(indices[[1]])
-    lines <- paste0(silver(indices), lines)
-  } else {
-    style <- style %||% cli_box_chars()
-    lines <- paste0(" ", style$h, lines)
-  }
+  indices <- tree$id
+  indices <- pad_spaces(as.character(indices))
+  indices <- paste0(" ", indices, ". ")
+  padding <- spaces(nchar(indices[[1]]))
+
+  lines <- paste0(silver(indices), lines)
+
+  src_locs <- tree$src_loc
+  src_locs <- map_if(src_locs, nzchar, ~ paste0(padding, "  at ", .x))
+  src_locs <- style_locs(src_locs)
+
+  lines <- zip_chr(lines, src_locs)
 
   if (is_null(max)) {
     return(lines)
@@ -390,11 +486,7 @@ cli_branch <- function(lines, max = NULL, style = NULL, indices = NULL) {
   style <- style %||% cli_box_chars()
   n_collapsed <- n - max
 
-  if (numbered) {
-    collapsed_line <- paste0(spaces(padding), "...")
-  } else {
-    collapsed_line <- format_collapsed_branch("...", n_collapsed, style = style)
-  }
+  collapsed_line <- paste0(padding, "...")
 
   if (max == 1L) {
     lines <- chr(
@@ -415,6 +507,19 @@ cli_branch <- function(lines, max = NULL, style = NULL, indices = NULL) {
   )
 }
 
+style_locs <- function(locs) {
+  chr(map_if(locs, nzchar, col_grey))
+}
+zip_chr <- function(xs, ys) {
+  flatten_chr(map2(xs, ys, function(x, y) {
+    if (nzchar(y)) {
+      c(x, y)
+    } else {
+      x
+    }
+  }))
+}
+
 
 #' @export
 print.rlang_trace <- function(x,
@@ -447,34 +552,6 @@ summary.rlang_trace <- function(object,
   invisible(object)
 }
 
-#' @rdname trace_back
-#' @param trace A backtrace created by `trace_back()`.
-#' @export
-trace_length <- function(trace) {
-  length(trace$calls)
-}
-
-trace_subset <- function(x, i) {
-  if (!length(i)) {
-    return(new_trace(list(), int()))
-  }
-  stopifnot(is_integerish(i))
-
-  n <- trace_length(x)
-
-  if (all(i < 0L)) {
-    i <- setdiff(seq_len(n), abs(i))
-  }
-
-  parents <- match(as.character(x$parents[i]), as.character(i), nomatch = 0)
-
-  new_trace(
-    calls = x$calls[i],
-    parents = parents,
-    indices = x$indices[i]
-  )
-}
-
 # Subsets sibling nodes, at the level of the rightmost leaf by
 # default. Supports full vector subsetting semantics (negative values,
 # missing index, etc).
@@ -483,12 +560,12 @@ trace_subset_across <- function(trace, i, n = NULL) {
   level_n <- length(level)
   i <- validate_index(i, level_n)
 
-  indices <- unlist(map(level[i], chain_indices, trace$parents))
-  trace_subset(trace, indices)
+  indices <- unlist(map(level[i], chain_indices, trace$parent))
+  trace_slice(trace, indices)
 }
 trace_level <- function(trace, n = NULL) {
   n <- n %||% trace_length(trace)
-  parents <- trace$parents
+  parents <- trace$parent
   which(parents == parents[[n]])
 }
 
@@ -530,38 +607,45 @@ parents_indices <- function(i, parents) {
 # Trimming ----------------------------------------------------------------
 
 trace_trim_env <- function(x, frames, to) {
+  idx <- trace_trim_env_idx(trace_length(x), frames, to)
+  trace_slice(x, idx)
+}
+trace_trim_env_idx <- function(n, frames, to) {
   to <- to %||% peek_option("rlang_trace_top_env")
 
+  # Trim knitr context if available
+  if (is_null(to) && is_true(peek_option('knitr.in.progress'))) {
+    to <- knitr::knit_global()
+  }
+
   if (is.null(to)) {
-    return(x)
+    return(TRUE)
   }
 
   is_top <- map_lgl(frames, is_reference, to)
   if (!any(is_top)) {
-    return(x)
+    return(TRUE)
   }
 
   start <- last(which(is_top)) + 1L
-  end <- trace_length(x)
-
-  trace_subset(x, seq2(start, end))
+  seq2(start, n)
 }
 
 set_trace_skipped <- function(trace, id, n) {
-  attr(trace$calls[[id]], "collapsed") <- n
+  trace$collapsed[[id]] <- n
   trace
 }
 set_trace_collapsed <- function(trace, id, n) {
-  attr(trace$calls[[id - n]], "collapsed") <- n
+  trace$collapsed[[id - n]] <- n
   trace
 }
 n_collapsed <- function(trace, id) {
-  call <- trace$calls[[id]]
+  call <- trace$call[[id]]
 
   if (is_eval_call(call)) {
     # When error occurs inside eval()'s frame at top level, there
     # might be only one frame and nothing to collapse
-    if (id > 1L && is_eval_call(trace$calls[[id - 1L]])) {
+    if (id > 1L && is_eval_call(trace$call[[id - 1L]])) {
       n <- 1L
     } else {
       n <- 0L
@@ -584,111 +668,24 @@ is_eval_call <- function(call) {
   is_call2(call, c("eval", "evalq"), ns = c("", "base"))
 }
 
-pipe_collect_calls <- function(pipe, env) {
-  node <- node_cdr(pipe)
-
-  last_call <- pipe_add_dot(node_cadr(node))
-  last_call <- maybe_add_namespace(last_call, env)
-
-  calls <- new_node(last_call, NULL)
-
-  while (is_call2(node_car(node), "%>%")) {
-    node <- node_cdar(node)
-    call <- pipe_add_dot(node_cadr(node))
-    call <- maybe_add_namespace(call, env)
-    calls <- new_node(call, calls)
-  }
-
-  first_call <- node_car(node)
-  if (is_call2(first_call)) {
-    # The first call doesn't need a dot
-    first_call <- maybe_add_namespace(first_call, env)
-    calls <- new_node(first_call, calls)
-    leading <- TRUE
-  } else {
-    leading <- FALSE
-  }
-
-  list(calls = as.list(calls), leading = leading)
-}
-pipe_add_dot <- function(call) {
-  if (!is_call2(call)) {
-    return(call2(call, dot_sym))
-  }
-
-  node <- node_cdr(call)
-  while (!is_null(node)) {
-    if (identical(node_car(node), dot_sym)) {
-      return(call)
-    }
-    node <- node_cdr(node)
-  }
-
-  args <- new_node(dot_sym, node_cdr(call))
-  new_call(node_car(call), args)
-}
-
-has_pipe_pointer <- function(x) {
-  !is_null(attr(x, "pipe_pointer"))
-}
-
-# Assumes a backtrace branch with collapsed pipe
-branch_uncollapse_pipe <- function(trace) {
-  orig <- trace
-
-  while (idx <- detect_index(trace$calls, has_pipe_pointer)) {
-    trace_before <- trace_subset(trace, seq2(1L, idx - 1L))
-    trace_after <- trace_subset(trace, seq2(idx + 2L, trace_length(trace)))
-
-    pipe <- trace$calls[[idx]]
-
-    pointer <- attr(pipe, "pipe_pointer")
-    if (!is_scalar_integer(pointer)) {
-      stop("Internal error: Invalid pipe pointer")
-    }
-
-    pipe_info <- attr(pipe, "pipe_info")
-    pipe_calls <- pipe_info$calls
-
-    if (pipe_info$leading) {
-      pointer <- inc(pointer)
-    }
-
-    incomplete <- seq2(pointer + 1L, length(pipe_calls))
-    if (length(incomplete)) {
-      pipe_calls <- pipe_calls[-incomplete]
-    }
-
-    parent <- trace$parents[[idx]]
-    pipe_parents <- seq(parent, parent + pointer - 1L)
-
-    # This assigns the same frame number to all pipe calls that have
-    # already returned
-    n_collapsed <- attr(pipe, "collapsed") %||% 0L
-    pipe_indices <- rep(idx, length(pipe_calls) - 1L)
-    pipe_indices <- c(pipe_indices, idx + n_collapsed + 1L)
-
-    # Add the number of uncollapsed frames to children's
-    # ancestry. This assumes a backtrace branch.
-    trace_after$parents <- trace_after$parents + pointer
-
-    trace$calls <- c(trace_before$calls, pipe_calls, trace_after$calls)
-    trace$parents <- c(trace_before$parents, pipe_parents, trace_after$parents)
-    trace$indices <- c(trace_before$indices, pipe_indices, trace_after$indices)
-  }
-
-  if (length(unique(lengths(trace))) != 1L) {
-    warn("Internal error: Trace data is not square.")
-    return(orig)
-  }
-
-  trace
-}
-
 trace_simplify_branch <- function(trace) {
-  parents <- trace$parents
-  path <- int()
-  id <- length(parents)
+  if (!trace_length(trace)) {
+    return(trace)
+  }
+
+  parents <- trace$parent
+
+  old_visible <- trace$visible
+  visible <- rep_along(old_visible, FALSE)
+
+  old_visible_loc <- which(old_visible)
+  if (length(old_visible_loc)) {
+    id <- last(old_visible_loc)
+  } else {
+    id <- 0L
+  }
+
+  trace$collapsed <- 0L
 
   while (id != 0L) {
     n_collapsed <- n_collapsed(trace, id)
@@ -696,28 +693,30 @@ trace_simplify_branch <- function(trace) {
     if (n_collapsed) {
       trace <- set_trace_collapsed(trace, id, n_collapsed)
       next_id <- id - n_collapsed
-
-      # Rechain child of collapsed parent to correct parent
-      parents[[id + 1L]] <- next_id
-
       id <- next_id
     }
 
-    if (!is_uninformative_call(trace$calls[[id]])) {
-      path <- c(path, id)
+    # Set `old_visible` to avoid uninformative calls in position 1 to
+    # be included (see below)
+    if (is_uninformative_call(trace$call[[id]])) {
+      old_visible[[id]] <- FALSE
+    } else if (old_visible[[id]]) {
+      visible[[id]] <- TRUE
     }
 
     id <- parents[id]
   }
 
-  # Always include very first call
-  path <- rev(path)
-  if (length(path) && path[[1]] != 1L) {
-    path <- c(1L, path)
+  # Always include first visible call
+  first <- detect_index(old_visible, is_true)
+  if (first) {
+    visible[[first]] <- TRUE
   }
 
-  trace$parents <- parents
-  trace_subset(trace, path)
+  trace$visible <- visible
+  trace$parent <- parents
+
+  trace
 }
 
 # Bypass calls with inlined functions
@@ -765,9 +764,18 @@ is_winch_frame <- function(call) {
 }
 
 trace_simplify_collapse <- function(trace) {
-  parents <- trace$parents
-  path <- int()
-  id <- length(parents)
+  if (!trace_length(trace)) {
+    return(trace)
+  }
+
+  parents <- trace$parent
+
+  old_visible <- trace$visible
+  visible <- rep_along(old_visible, FALSE)
+
+  id <- trace_length(trace)
+
+  trace$collapsed <- 0L
 
   while (id > 0L) {
     n_collapsed <- n_collapsed(trace, id)
@@ -782,7 +790,7 @@ trace_simplify_collapse <- function(trace) {
       id <- next_id
     }
 
-    path <- c(path, id)
+    visible[[id]] <- TRUE
     parent_id <- parents[[id]]
     id <- dec(id)
 
@@ -793,7 +801,7 @@ trace_simplify_collapse <- function(trace) {
 
       if (sibling_parent_id == parent_id) {
         trace <- set_trace_skipped(trace, id, n_skipped)
-        path <- c(path, id)
+        visible[[id]] <- TRUE
         n_skipped <- 0L
       } else {
         n_skipped <- inc(n_skipped)
@@ -803,62 +811,109 @@ trace_simplify_collapse <- function(trace) {
     }
   }
 
-  trace$parents <- parents
-  trace_subset(trace, rev(path))
+  trace$visible <- visible & old_visible
+  trace$parent <- parents
+
+  trace
 }
 
 
 # Printing ----------------------------------------------------------------
 
-trace_as_tree <- function(x, dir = getwd(), srcrefs = NULL) {
-  nodes <- c(0, seq_along(x$calls))
-  children <- map(nodes, function(id) seq_along(x$parents)[x$parents == id])
+trace_as_tree <- function(trace, dir = getwd(), srcrefs = NULL) {
+  if (is_null(trace$collapsed)) {
+    trace$collapsed <- vec_recycle(0L, trace_length(trace))
+  }
+  call_text_data <- trace[c("call", "collapsed", "namespace", "scope")]
+  call_text <- chr(!!!pmap(call_text_data, trace_call_text))
 
-  calls <- as.list(x$calls)
-  is_collapsed <- map(calls, attr, "collapsed")
-  call_text <- map2_chr(calls, is_collapsed, trace_call_text)
-
-  srcrefs <- srcrefs %||% peek_option("rlang_trace_format_srcrefs")
-  srcrefs <- srcrefs %||% TRUE
+  srcrefs <- srcrefs %||% peek_option("rlang_trace_format_srcrefs") %||% TRUE
   stopifnot(is_scalar_logical(srcrefs))
+
   if (srcrefs) {
-    refs <- map(x$calls, attr, "srcref")
-    src_locs <- map_chr(refs, src_loc, dir = dir)
-    have_src_loc <- nzchar(src_locs)
-    src_locs <- silver(src_locs[have_src_loc])
-    call_text[have_src_loc] <- paste0(call_text[have_src_loc], " ", src_locs)
+    refs <- map(trace$call, attr, "srcref")
+    src_locs <- map_chr(refs, src_loc)
+    trace$src_loc <- src_locs
+  } else {
+    trace$src_loc <- vec_recycle("", trace_length(trace))
   }
 
-  tree <- data.frame(id = as.character(nodes), stringsAsFactors = FALSE)
-  tree$children <- map(children, as.character)
-  tree$call <- c(trace_root(), call_text)
+  id <- c(0, seq_along(trace$call))
+  children <- map(id, function(id) seq_along(trace$parent)[trace$parent == id])
+
+  root <- data_frame(
+    call = list(NULL),
+    parent = 0L,
+    visible = TRUE,
+    namespace = NA,
+    scope = NA,
+    src_loc = ""
+  )
+  trace <- vec_rbind(root, trace)
+
+  tree_data <- data_frame(
+    id = as.character(id),
+    children = children,
+    call_text = c(trace_root(), call_text)
+  )
+  tree <- vec_cbind(tree_data, trace)
+
+  # Subset out hidden frames
+  tree <- vec_slice(tree, tree$visible)
+  tree$children <- map(tree$children, intersect, tree$id)
+
+  if (has_crayon()) {
+    # Detect runs of namespaces/global
+    ns <- tree$namespace
+    ns <- ifelse(is.na(ns) & tree$scope == "global", "global", ns)
+    ns[[1]] <- "_root"
+    starts <- detect_run_starts(ns)
+
+    # Embolden first occurrences in runs of namespaces/global
+    tree$call_text <- map2_chr(tree$call_text, starts, function(text, start) {
+      if (is_true(start)) {
+        text <- sub(
+          "^([a-zA-Z]+)(::|:::| )",
+          sprintf("%s\\1%s\\2", open_bold(), close_bold()),
+          text
+        )
+      }
+      text
+    })
+  }
 
   tree
 }
 
 # FIXME: Add something like call_deparse_line()
-trace_call_text <- function(call, collapse) {
-  if (is_null(collapse)) {
-    return(as_label(call))
+trace_call_text <- function(call, collapsed, namespace, scope) {
+  if (collapsed && length(call) > 1L) {
+    call <- call2(call[[1]], quote(...))
   }
 
-  if (is_call2(call, "%>%")) {
-    call <- call
-  } else if (length(call) > 1L) {
-    call <- call2(node_car(call), quote(...))
+  if (is_call(call) && is_symbol(call[[1]])) {
+    if (scope %in% c("::", ":::") && !is_na(namespace)) {
+      call[[1]] <- call(scope, sym(namespace), call[[1]])
+    }
   }
 
   text <- as_label(call)
-  if (collapse > 0L) {
-    n_collapsed_text <- sprintf(" ... +%d", collapse)
-  } else {
-    n_collapsed_text <- ""
+
+  if (is_string(scope, "global")) {
+    text <- paste0("global ", text)
+  } else if (is_string(scope, "local") && !is_na(namespace)) {
+    text <- paste0(namespace, " ", text)
   }
 
-  format_collapsed(paste0("[ ", text, " ]"), collapse)
+  if (collapsed) {
+    n_collapsed_text <- sprintf(" ... +%d", collapsed)
+    text <- format_collapsed(paste0("[ ", text, " ]"), collapsed)
+  }
+
+  text
 }
 
-src_loc <- function(srcref, dir = getwd()) {
+src_loc <- function(srcref) {
   if (is.null(srcref)) {
     return("")
   }
@@ -872,26 +927,20 @@ src_loc <- function(srcref, dir = getwd()) {
   if (identical(file, "") || identical(file, "<text>")) {
     return("")
   }
-  if (!file.exists(file) && is_null(peek_option("rlang:::trace_force_dangling_srcrefs"))) {
-    return("")
+
+  if (is_null(peek_option("rlang:::disable_trim_srcref"))) {
+    file <- path_trim_prefix(file, 3)
   }
 
   line <- srcref[[1]]
   column <- srcref[[5]] - 1L
-  paste0(relish(file, dir = dir), ":", line, ":", column)
-}
 
-relish <- function(x, dir = getwd()) {
-  if (substr(dir, nchar(dir), nchar(dir)) != "/") {
-    dir <- paste0(dir, "/")
-  }
-
-  gsub(dir, "", x, fixed = TRUE)
+  paste0(file, ":", line, ":", column)
 }
 
 trace_root <- function() {
   if (cli_is_utf8_output()) {
-    "\u2588"
+    "\u2586"
   } else {
     "x"
   }
@@ -909,7 +958,7 @@ trace_root <- function() {
 #' @author Iñaki Ucar (ORCID: 0000-0001-6403-5550)
 #' @noRd
 trace_pkgs <- function(pkgs, max_level = Inf, ..., regexp = NULL) {
-  check_dots_empty(...)
+  check_dots_empty()
 
   # Avoids namespace loading issues
   lapply(pkgs, requireNamespace, quietly = TRUE)
@@ -1000,7 +1049,7 @@ call_add_namespace <- function(call, fn) {
     return(call)
   }
 
-  namespaced_sym <- call(op, as.symbol(prefix), sym)
+  namespaced_sym <- call(op, sym(prefix), sym)
   call[[1]] <- namespaced_sym
   call
 }
@@ -1008,3 +1057,54 @@ call_add_namespace <- function(call, fn) {
 is_trace <- function(x) {
   inherits_any(x, c("rlang_trace", "rlib_trace"))
 }
+
+#' Backtrace specification
+#'
+#' @description
+#'
+#' `r lifecycle::badge("experimental")`
+#'
+#'
+#' @section Structure:
+#'
+#' An r-lib backtrace is a data frame that contains the following
+#' columns:
+#'
+#' - `call`: List of calls. These may carry `srcref` objects.
+#'
+#' - `visible`: Logical vector. If `FALSE`, the corresponding call
+#'   will be hidden from simplified backtraces.
+#'
+#' - `parent`: Integer vector of parent references (see
+#'   [sys.parents()]) as row numbers. 0 is global.
+#'
+#' - `namespace`: Character vector of namespaces. `NA` for global or
+#'   no namespace
+#'
+#' - `scope`: Character vector of strings taking values `"::"`,
+#'   `":::"`, `"global"`, or `"local"`.
+#'
+#' A backtrace data frame may contain extra columns. If you add
+#' additional columns, make sure to prefix their names with the name
+#' of your package or organisation to avoid potential conflicts with
+#' future extensions of this spec, e.g. `"mypkg_column"`.
+#'
+#'
+#' @section Operations:
+#'
+#' - **Length**. The length of the backtrace is the number of rows of
+#'   the underlying data.
+#'
+#' - **Concatenation**. Performed by row-binding two backtraces.  The
+#'   `parent` column of the RHS is shifted by `nrow(LHS)` so that the
+#'   last call of the LHS takes place of the global frame of the RHS.
+#'
+#' - **Subsetting**. Performed by slicing the backtrace. After the
+#'   data frame is sliced, the `parent` column is adjusted to the new
+#'   row indices. Any `parent` value that no longer exists in the
+#'   sliced backtrace is set to 0 (the global frame).
+#'
+#'
+#' @name rlib_trace_spec
+#' @keywords internal
+NULL
